package jmaf.core;

import java.io.File;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.Calendar;

import org.eclipse.core.runtime.ILog;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;

public final class CommonTools {
	private static final String PLUGIN_ID = "foremka3";
	private static ILog log;

	/**
	 * Returns file extension
	 * @param file Given file
	 * @return String containing file extension or empty string if file 
	 * has no extension
	 */
	public static String getExtension(File file) {
		int idx = file.getName().lastIndexOf('.');
		if (idx == -1) {
			return "";
		}
		return file.getName().substring(
				 idx + 1);
	}
	
	/**
	 * Returns current time formated like hh:mm:ss
	 * @return current time formated like hh:mm:ss
	 */
	public static String getTime() {
		Calendar c = Calendar.getInstance();
		return addZero(c.get(Calendar.HOUR_OF_DAY)) + ":" +
			addZero(c.get(Calendar.MINUTE)) + ":" +
			addZero(c.get(Calendar.SECOND));
	}
	
	/**
	 * 
	 * @return Current time that can be a part of a file name
	 */
	public static String getFileTime() {
		Calendar c = Calendar.getInstance();
		return addZero(c.get(Calendar.HOUR_OF_DAY)) + "-" +
			addZero(c.get(Calendar.MINUTE)) + "-" +
			addZero(c.get(Calendar.SECOND));
	}
	
	/**
	 * If given numer has only one sign add '0'
	 * sign at the begining and retruns a string
	 * @param input Input number
	 * @return String with added 0 if needed
	 */
	private static String addZero(int number){
		String str = Integer.toString(number);
		if (str.length()==1){
			return "0"+str;
		}
		return str;
	}
	/**
	 * Prints message and new line separator
	 * 
	 * @param out Output stream to put message on
	 * @param message Message to puts
	 */
	public static void println(OutputStreamWriter out, String message) {
		try {
			out.write(message + System.getProperty("line.separator"));
			out.flush();
		} catch (IOException e1) {
			e1.printStackTrace();
		}
	}
	
//	public static void listAlgorithms(){
//		IExtensionRegistry extReg = Platform.getExtensionRegistry();
//		System.out.println("2______________________________________________________");
//		IExtensionPoint extensionPoint = extReg.getExtensionPoint("ForemkaCore.algorithm");
//		IExtension[] ePExtensions = extensionPoint.getExtensions();
//		System.out.println(ePExtensions.length + " extensions of "+ "ForemkaCore.algorithm");
//		for (int i = 0; i < ePExtensions.length; i++) {
//			System.out.println("Ext "+i + ": "+ ePExtensions[i].getLabel());
//		}
//		
//		System.out.println("______________________________________________________");
//	}

	public static void setLog(ILog log) {
		CommonTools.log = log;
	}
	
	public static void log(String message){
		log(message, null);
	}
	public static void log(String message, Throwable exception){
		log.log(new Status(IStatus.ERROR, PLUGIN_ID, message, exception));
		if (exception!=null){
			exception.printStackTrace();
		}
		if (message!=null){
			System.err.println(message);
		}
	}

	public static void log(Throwable t) {
		log((t.getMessage()==null) ? null : t.getMessage(), t);
	}

	/**
	 * Removes extension in file name if one exists
	 * 
	 * @param fileName Given file name
	 * @return File name without extension
	 */
	public static String removeExtension(String fileName) {
		int index = fileName.lastIndexOf('.');
		if (index>0){
			fileName = fileName.substring(0, index);
		}
		return fileName;
	}
	
}
