package jmaf.core;

import java.io.BufferedReader;
import java.io.File;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.List;

import jmaf.core.exceptions.CannotCalculateException;


/**
 * Abstract class supporting running external file algorithms
 * @author Micha
 *
 */
public abstract class ExternalFileAlgorithm implements IAlgorithm{
	/**
	 * Returns a path to external file
	 * @return Path to external file
	 */
	public abstract String getExternalFile(); 
	
	/**
	 * Initial directory of the external file execution. If null, current 
	 * directory is selected
	 *  
	 * @return initial directory of the external file execution
	 */
	public abstract String getInitialDirectory();
	
	/**
	 * 
	 * @param inputFile Input file reference
	 * @param outputFile Output file reference
	 * @param parameters List of validated parameters values (pairs id, value)
	 * @return List<String> List of commands parameters that will be passed to
	 * 	algorithm through command line
	 */
	public abstract List<String> mapParametersToArguments(File inputFile, File outputFile, List<String[]> parameters);
	
	public int performComputation(File inputFile, File outputFile, List<String[]> parameters, OutputStreamWriter outputStream, IProgressNotifier progressMonitor) {
		OutputStreamWriter out = outputStream;

		List<String> command = new ArrayList<String>();
		String externalFile = getExternalFile();
		if (!new File(externalFile).exists()){
			throw new CannotCalculateException("File '"+externalFile+"' doesn't exist.");
		}
	    command.add(externalFile);    
	    
	    List<String> arguments = mapParametersToArguments(inputFile, outputFile, parameters);
	    if (arguments!=null){
	    	for (String str : arguments) {
				command.add(str);
			}
	    }

	    ProcessBuilder builder = new ProcessBuilder(command);
		 
	    String dir = getInitialDirectory();
	    if (dir==null){
	    	dir = ".";
	    }

	    builder.directory(new File(dir));
	    
	    builder.redirectErrorStream(true);

        try {
	        Process process = builder.start();

	        InputStream is = process.getInputStream();
	        InputStreamReader isr = new InputStreamReader(is);
	        BufferedReader br = new BufferedReader(isr);
	        String line;

	        while ((line = br.readLine()) != null) {
	          CommonTools.println(out, line);
	        }
	        
        } catch (Throwable t){
        	CommonTools.println(out, t.getMessage());
        	return IAlgorithm.STATUS_FATAL_ERROR;
       }
        
		return IAlgorithm.STATUS_OK;
	}

	
}
