package jmaf.core;

import java.io.File;
import java.io.OutputStreamWriter;
import java.util.List;

/**
 * Interface for performing cross validation 
 * 
 * @author Michal Janiszewski
 *
 */
public interface ICrossValidationAlgorithm {
	/**
	 * Performs cross validation
	 * @param inputFile Input file
	 * @param outputDirectory Directory where output files should be stored
	 * @param parameters List of parameters
	 * @param outputStreamWriter 
	 * @param rwp
	 * @return object of type <code>ClassificationResultsValidationContainer</code>
	 */
	public Object performCrossValidation(File inputFile, File outputDirectory, List<String[]> parameters,
			OutputStreamWriter outputStreamWriter, IProgressNotifier rwp);
	/**
	 * Performs validation of cross validation parameter
	 * 
	 * @param parameterID Parameter identifier
	 * @param value Parameter value
	 * @return <code>null</code> if valid parameter value, message describing reason
	 * or error otherwise
	 */
	public String validateCrossValidationParameter(String parameterID, String value);

}
